﻿--[[************************************************************************************************************************************************************
	
	multiTrailer v1.5.5
	
	"Farming Simulator 2013" & "Farming Simulator 2013 MORE REALISTIC" & "Farming Simulator 2015"
	
	06/2015 © Saty MODS (Saty / http://forum.lsczech.cz)
	
****************************************************************************************************************************************************************
	
	(EN) INFO
	
	Free for use on mods. Modifications only with my permission !
	
	This work is licensed under a Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International License (Attribution / NonCommercial / NoDerivatives)
	http://creativecommons.org/licenses/by-nc-nd/4.0/deed.en
	
****************************************************************************************************************************************************************
	
	(CZ) INFO
	
	Pro volne pouziti v modech. ZAKAZ modifikaci bez meho souhlasu !
	
	Toto dilo podleha licenci : Creative Commons 4.0 International - Uvedte autora / Neuzivejte dilo komercne / Nezasahujte do dila 
	http://creativecommons.org/licenses/by-nc-nd/4.0/deed.cs
	
************************************************************************************************************************************************************]]--

source(g_currentModDirectory .. "Scripts/multiTrailerEvent.lua");  -- pokud nejsou skripty umistene ve slozce "scripts" tak zde upravit cestu !!
multiTrailer = {};

function multiTrailer.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Fillable, specializations) and SpecializationUtil.hasSpecialization(Trailer, specializations);
end;

function multiTrailer:load(xmlFile)
	self.changeCurrentBody = SpecializationUtil.callSpecializationsFunction("changeCurrentBody");
	self.setCoverSheet = SpecializationUtil.callSpecializationsFunction("setCoverSheet");
	self.setOpenableSidewalls = SpecializationUtil.callSpecializationsFunction("setOpenableSidewalls");
	self.setUnloadingObjects = SpecializationUtil.callSpecializationsFunction("setUnloadingObjects");
	self.isActiveForInput = multiTrailer.isActiveForInput;
	
	self.multiT = {};
	self.multiT.body = {};
	
	self.multiT.firstLoad = true;
	self.multiT.onlyRepair = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.multiTrailer#onlyInRepairTrigger"), false);
	self.multiT.activeBody = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.multiTrailer#defaultBody"), 1);
	self.multiT.realCapacity = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.multiTrailer#realCapacity"), true);
	self.multiT.realMaxLoad = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.multiTrailer#maximalLoad"), 0);
	local useRealFillPlanes = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.multiTrailer#realFillPlanes"), true);
	
	local coverSheetTriggerNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.multiTrailer#coverSheetTriggerNode"));
	if coverSheetTriggerNode ~= nil then
		self.coverSheetTrigger = multiTrailer.coverSheetTrigger;
		self.multiT.coverSheetTriggerNode = coverSheetTriggerNode;
		addTrigger(self.multiT.coverSheetTriggerNode, "coverSheetTrigger", self);
		self.multiT.coverSheetInRange = false;
	end;
	
	-- ## OVERRIDE BRAKES WITH MAXIMAL LOAD ##
	if self.multiT.realMaxLoad > 0 and self.isRealistic then
		self.multiT.realCropLoad = 0;
		self.multiT.balesMass = 0;
		self.multiT.brakeDecrease = 0;
		local i = 0;
		while true do
			local wheelnamei = string.format("vehicle.wheels.wheel(%d)", i);
			if self.wheels[i+1] == nil then
				break;
			end;
			self.wheels[i+1].brakeRatioBackup = Utils.getNoNil(getXMLFloat(xmlFile, wheelnamei .. "#brakeRatio"), 0);
			i = i + 1;
		end;
	end;
	
	-- ## REALISTIC BALE SEARCH TRIGGER ##
	local realisticBaleSearchTriggerNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.multiTrailer#realisticBaleSearchTrigger"));
	if realisticBaleSearchTriggerNode ~= nil then
		self.realisticBaleSearchTrigger = multiTrailer.realisticBaleSearchTrigger;
		self.multiT.realisticBaleSearchTriggerNode = realisticBaleSearchTriggerNode;
		addTrigger(self.multiT.realisticBaleSearchTriggerNode, "realisticBaleSearchTrigger", self);
		self.multiT.numBalesInTrigger = 0;
		self.multiT.numVehiclesInTrigger = 0;
	end;
	
	-- ## OPENABLE SIDEWALLS ##
	self.multiT.sidewallsChain = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.multiTrailer.openableSidewalls#chainNode"));
	local leftKey = "vehicle.multiTrailer.openableSidewalls.leftSidewall";
	local leftAnimation = getXMLString(xmlFile, leftKey .. "#animation");
	local leftTriggerNode = Utils.indexToObject(self.components, getXMLString(xmlFile, leftKey .. "#triggerNode"));
	if leftTriggerNode ~= nil and self.animations ~= nil and self.animations[leftAnimation] ~= nil then
		self.leftTrigger = multiTrailer.leftTrigger;
		self.multiT.leftTriggerNode = leftTriggerNode;
		addTrigger(self.multiT.leftTriggerNode, "leftTrigger", self);
		self.multiT.leftSidewall = {};
		self.multiT.leftSidewall.inRange = false;
		self.multiT.leftSidewall.animation = leftAnimation;
		self.multiT.leftSidewall.isOpen = false;
	end;
	local rightKey = "vehicle.multiTrailer.openableSidewalls.rightSidewall";
	local rightAnimation = getXMLString(xmlFile, rightKey .. "#animation");
	local rightTriggerNode = Utils.indexToObject(self.components, getXMLString(xmlFile, rightKey .. "#triggerNode"));
	if rightTriggerNode ~= nil and self.animations ~= nil and self.animations[rightAnimation] ~= nil then
		self.rightTrigger = multiTrailer.rightTrigger;
		self.multiT.rightTriggerNode = rightTriggerNode;
		addTrigger(self.multiT.rightTriggerNode, "rightTrigger", self);
		self.multiT.rightSidewall = {};
		self.multiT.rightSidewall.inRange = false;
		self.multiT.rightSidewall.animation = rightAnimation;
		self.multiT.rightSidewall.isOpen = false;
	end;
	
	-- ## BODY TYPES ##
	local i = 0;
	while true do
		local key = string.format("vehicle.multiTrailer.body(%d)", i);
		if not hasXMLProperty(xmlFile, key) then
			break;
		end;
		local capacity = getXMLInt(xmlFile, key .. "#capacity");
		local allowFillFromAir = Utils.getNoNil(getXMLBool(xmlFile, key .. "#allowFillFromAir"), true);
		local dynamicMountAttacher = Utils.getNoNil(getXMLBool(xmlFile, key .. "#dynamicMountAttacher"), not self.isRealistic);
		local leftSidewallActive = Utils.getNoNil(getXMLBool(xmlFile, key .. "#leftSidewall"), false);
		local rightSidewallActive = Utils.getNoNil(getXMLBool(xmlFile, key .. "#rightSidewall"), false);
		
		-- ## FILL TYPES ##
		local fillTypes = {};
		local fillTypesLoad = getXMLString(xmlFile, key .. ".supportedFillTypes#fillTypes");
		if fillTypesLoad ~= nil then
			local types = Utils.splitString(" ", fillTypesLoad);
			for k,v in pairs(types) do
				local fillType = Fillable.fillTypeNameToInt[v];
				if fillType ~= nil then
					fillTypes[fillType] = true;
					if self.fillTypes[fillType] == nil then
						self.fillTypes[fillType] = true;
					end;
				end;
			end;
		end;
		
		-- ## COMPONENTS WITH CHANGES VISIBILITY ##
		local components = {};
		local j = 0;
		while true do
			local keyNode = key .. string.format(".components.component(%d)", j);
			if not hasXMLProperty(xmlFile, keyNode) then
				break;
			end;
			local node = Utils.indexToObject(self.components, getXMLString(xmlFile, keyNode.."#node"));
			table.insert(components, node);
			j = j + 1;
		end;
		
		-- ## COVERING SHEET ##
		self.multiT.coverSheet = false;
		local coverSheetUnused = getXMLString(xmlFile, key .. "#coverSheetUnused");
		if coverSheetUnused ~= nil then
			coverSheetUnused = Utils.indexToObject(self.components, coverSheetUnused);
			setVisibility(coverSheetUnused, false);
		end;
		
		local coverSheet = getXMLString(xmlFile, key .. "#coverSheetNode");
		if coverSheet ~= nil then
			coverSheet = Utils.indexToObject(self.components, coverSheet);
			setVisibility(coverSheet, false);
			
			local coverSheetScale = {};
			local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. ".coverSheetScalePart#node"));
			if node ~= nil then
				local defaultX, defaultY, defaultZ = getTranslation(node);
				local defaultRX, defaultRY, defaultRZ = getRotation(node);
				local animCurve = AnimCurve:new(linearInterpolatorTransRotScale);
				local keyI = 0;
				while true do
					local animKey = key .. string.format(".coverSheetScalePart.key(%d)", keyI);
					local keyTime = getXMLFloat(xmlFile, animKey.."#time");
					local x,y,z = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#translation"));
					if y == nil then
						y = getXMLFloat(xmlFile, animKey.."#y");
					end;
					local rx,ry,rz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#rotation"));
					local sx,sy,sz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#scale"));
					if keyTime == nil then
						break;
					end;
					local x = Utils.getNoNil(x, defaultX);
					local y = Utils.getNoNil(y, defaultY);
					local z = Utils.getNoNil(z, defaultZ);
					local rx = Utils.getNoNil(rx, defaultRX);
					local ry = Utils.getNoNil(ry, defaultRY);
					local rz = Utils.getNoNil(rz, defaultRZ);
					local sx = Utils.getNoNil(sx, 1);
					local sy = Utils.getNoNil(sy, 1);
					local sz = Utils.getNoNil(sz, 1);
					animCurve:addKeyframe({x=x, y=y, z=z, rx=rx, ry=ry, rz=rz, sx=sx, sy=sy, sz=sz, time = keyTime});
					keyI = keyI + 1;
				end;
				if keyI == 0 then
					local minY, maxY = Utils.getVectorFromString(getXMLString(xmlFile, nodeKey.."#minMaxY"));
					local minY = Utils.getNoNil(minY, defaultY);
					local maxY = Utils.getNoNil(maxY, defaultY);
					animCurve:addKeyframe({x=defaultX, y=minY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 0});
					animCurve:addKeyframe({x=defaultX, y=maxY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 1});
				end;
				coverSheetScaleNode = node;
				coverSheetScaleAnimCurve = animCurve;
			end;
		end;
		
		-- ## REAL CAPACITY MULTIPLIER ##
		local realCapacityFx = {};
		local l = 0;
		while true do
			local keyRCM = key .. string.format(".realCapacityMultipliers.realCapacityMultiplier(%d)", l);
			if not hasXMLProperty(xmlFile, keyRCM) then
				break;
			end;
			local fillTypeString = getXMLString(xmlFile, keyRCM.."#fillType");
			local multiplier = getXMLFloat(xmlFile, keyRCM.."#multiplier");
			local fillType = Fillable.fillTypeNameToInt[fillTypeString];
			if fillType ~= nil then
				realCapacityFx[fillType] = multiplier;
			end;
			l = l + 1;
		end;
		
		-- ## FILL PLANES ##
		self.spareFillPlanes = {};
		local fillPlanes = {};
		local k = 0;
		while true do
			local keyPlanes = key .. string.format(".fillPlanes.fillPlane(%d)", k);
			if not hasXMLProperty(xmlFile, keyPlanes) then
				break;
			end;
			local fillPlane = {};
			fillPlane.nodes = {};
			local fillTypes = getXMLString(xmlFile, keyPlanes .. "#type");
			if fillTypes ~= nil then
				local types = Utils.splitString(" ", fillTypes);
				for _,fillType in pairs(types) do
					local fillTypeInt = Fillable.fillTypeNameToInt[fillType];
					local nodeI = 0;
					while true do
						local nodeKey = keyPlanes..string.format(".node(%d)", nodeI);
						if not hasXMLProperty(xmlFile, nodeKey) then
							 break;
						end;
						local node = Utils.indexToObject(self.components, getXMLString(xmlFile, nodeKey.."#index"));
						if node ~= nil then
							local defaultX, defaultY, defaultZ = getTranslation(node);
							local defaultRX, defaultRY, defaultRZ = getRotation(node);
							setVisibility(node, false);

							local animCurve = AnimCurve:new(linearInterpolatorTransRotScale);
							local keyI = 0;
							while true do
								local animKey = nodeKey..string.format(".key(%d)", keyI);
								local keyTime = getXMLFloat(xmlFile, animKey.."#time");
								local x,y,z = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#translation"));
								if x == nil then
									x = getXMLFloat(xmlFile, animKey.."#x");
								end;
								if y == nil then
									y = getXMLFloat(xmlFile, animKey.."#y");
								end;
								if z == nil then
									z = getXMLFloat(xmlFile, animKey.."#z");
								end;
								local rx,ry,rz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#rotation"));
								local sx,sy,sz = Utils.getVectorFromString(getXMLString(xmlFile, animKey.."#scale"));
								if keyTime == nil then
									break;
								end;
								if useRealFillPlanes and (self.isRealistic or self.multiT.realCapacity) and realCapacityFx[fillTypeInt] ~= nil and realCapacityFx[fillTypeInt] < 1 then
									sy = sy * realCapacityFx[fillTypeInt];
								end;
								local x = Utils.getNoNil(x, defaultX);
								local y = Utils.getNoNil(y, defaultY);
								local z = Utils.getNoNil(z, defaultZ);
								local rx = Utils.getNoNil(rx, defaultRX);
								local ry = Utils.getNoNil(ry, defaultRY);
								local rz = Utils.getNoNil(rz, defaultRZ);
								local sx = Utils.getNoNil(sx, 1);
								local sy = Utils.getNoNil(sy, 1);
								local sz = Utils.getNoNil(sz, 1);
								animCurve:addKeyframe({x=x, y=y, z=z, rx=rx, ry=ry, rz=rz, sx=sx, sy=sy, sz=sz, time = keyTime});
								keyI = keyI + 1;
							end;
							if keyI == 0 then
								local minX, maxX = Utils.getVectorFromString(getXMLString(xmlFile, nodeKey.."#minMaxX"));
								local minX = Utils.getNoNil(minX, defaultX);
								local maxX = Utils.getNoNil(maxX, defaultX);
								local minY, maxY = Utils.getVectorFromString(getXMLString(xmlFile, nodeKey.."#minMaxY"));
								local minY = Utils.getNoNil(minY, defaultY);
								local maxY = Utils.getNoNil(maxY, defaultY);
								local minZ, maxZ = Utils.getVectorFromString(getXMLString(xmlFile, nodeKey.."#minMaxZ"));
								local minZ = Utils.getNoNil(minZ, defaultZ);
								local maxZ = Utils.getNoNil(maxZ, defaultZ);
								animCurve:addKeyframe({x=minX, y=minY, z=minZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 0});
								animCurve:addKeyframe({x=maxX, y=maxY, z=maxZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 1});
							end;
							table.insert(fillPlane.nodes, {node=node, animCurve = animCurve});
						end;
						nodeI = nodeI + 1;
					end;
					fillPlanes[fillType] = fillPlane;
					if self.spareFillPlanes[fillType] == nil then
						self.spareFillPlanes[fillType] = fillPlane;
					end;
				end;
			end;
			k = k + 1;
		end;
		
		-- ## OPTIONAL TRAILER TIP REFERENCE POINTS ##
		local tipReferencePoints = {};
		local tipReferencePointsLoad = getXMLString(xmlFile, key .. "#usedTipReferencePoints");
		if tipReferencePointsLoad ~= nil then
			local used = Utils.splitString(" ", tipReferencePointsLoad);
			for i, j in pairs(used) do
				tipReferencePoints[i] = true;
			end;
		else
			for i, j in pairs(self.tipReferencePoints) do
				tipReferencePoints[i] = true;
			end;
		end;
		
		-- ## ONLY FOR USE IN LS15 ##
		local fillVolumes = {};
		local fillVolumeDeformers = {};
		local fillVolumeHeights = {};
		local fillVolumeHeightRefNodeToFillVolumeHeight = {};
		
		if self.fillVolumes ~= nil then
			-- ## FILLVOLUMES FOR LS15 ##
			local j = 0;
			while true do
				local keyNode = key .. string.format(".fillVolumes.fillVolume(%d)", j);
				if not hasXMLProperty(xmlFile, keyNode) then
					break;
				end;
				
				local fillVolume = {};
				
				fillVolume.baseNode = Utils.indexToObject(self.components, getXMLString(xmlFile, keyNode.."#index"));
				fillVolume.allSidePlanes = Utils.getNoNil(getXMLBool(xmlFile, keyNode.."#allSidePlanes"), true);
				
				local defaultFillType = getXMLString(xmlFile, keyNode.."#defaultFillType");
				if defaultFillType ~= nil then
					local fillType = Fillable.fillTypeNameToInt[defaultFillType];
					if fillType ~= nil then
						fillVolume.defaultFillType = fillType;
					else
						print("ERROR multiTrailer: Invalid defaultFillType '"..tostring(defaultFillType).."' in '"..self.configFileName.."'");
					end;
				end;
				
				fillVolume.maxDelta = Utils.getNoNil(getXMLFloat(xmlFile, keyNode.."#maxDelta"), 1.0);
				fillVolume.maxSurfaceAngle = math.rad( Utils.getNoNil(getXMLFloat(xmlFile, keyNode.."#maxAllowedHeapAngle"), 35) );
				
				local maxPhysicalSurfaceAngle = math.rad(35);
				fillVolume.maxSubDivEdgeLength = Utils.getNoNil(getXMLFloat(xmlFile, keyNode.."#maxSubDivEdgeLength"), 0.9);
				
				fillVolume.volume = createFillPlaneShape(fillVolume.baseNode, "fillPlane", capacity, fillVolume.maxDelta, fillVolume.maxSurfaceAngle, maxPhysicalSurfaceAngle, fillVolume.maxSubDivEdgeLength, fillVolume.allSidePlanes);
				
				fillVolume.deformers = {};
				if fillVolume.volume ~= nil then
					local k = 0;
					while true do
						local node = Utils.indexToObject(self.components, getXMLString(xmlFile, keyNode..".deformNode("..k..")#index"));
						if node == nil then
							break;
						end;
						local initPos = { localToLocal(node, fillVolume.baseNode, 0,0,0) };
						local polyline = findPolyline(fillVolume.volume, initPos[1],initPos[3]);
						fillVolumeDeformers[node] = {node=node, initPos=initPos, posX=initPos[1], posZ=initPos[3], polyline=polyline, volume=fillVolume.volume, baseNode=fillVolume.baseNode};
						k = k + 1;
					end;
				end;
				
				fillVolume.scrollSpeedDischarge = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, keyNode.."#scrollSpeedDischarge"), "0 0 0")) };
				fillVolume.scrollSpeedLoad = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, keyNode.."#scrollSpeedLoad"), "0 0 0")) };
				for i=1,3 do
					fillVolume.scrollSpeedDischarge[i] = fillVolume.scrollSpeedDischarge[i] / 1000;
					fillVolume.scrollSpeedLoad[i] = fillVolume.scrollSpeedLoad[i] / 1000;
				end;
				fillVolume.uvPosition = {0, 0, 0};
			
				if fillVolume.volume ~= nil and fillVolume.volume ~= 0 then
					link(fillVolume.baseNode, fillVolume.volume);
					table.insert(fillVolumes, fillVolume);
				end;
				
				j = j + 1;
			end;
			
			-- ## FILLVOLUMEHEIGHTS FOR LS15 ##
			local j = 0;
			while true do
				local keyNode = key ..  string.format(".fillVolumeHeights.fillVolumeHeight(%d)", j);
				if not hasXMLProperty(xmlFile, keyNode) then
					break;
				end;
				local volumeHeight = {};
				volumeHeight.fillVolumeId = getXMLInt(xmlFile, keyNode .. "#fillVolumeId");
				volumeHeight.volumeHeightIsDirty = false;

				volumeHeight.refNodes = {};
				local k = 0;
				while true do
					local refNode = Utils.indexToObject(self.components, getXMLString(xmlFile, string.format("%s.refNode(%d)#index", keyNode, k)));
					if refNode == nil then
						break;
					end;
					table.insert(volumeHeight.refNodes, {refNode=refNode});
					fillVolumeHeightRefNodeToFillVolumeHeight[refNode] = volumeHeight;
					k = k + 1;
				end;

				volumeHeight.nodes = {};
				local k = 0;
				while true do
					local node = Utils.indexToObject(self.components, getXMLString(xmlFile, string.format("%s.node(%d)#index", keyNode, k)));
					if node == nil then
						break;
					end;
					if node ~= nil then
						local baseScale = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#baseScale", keyNode, k)), "1 1 1")) };
						local scaleAxis = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#scaleAxis", keyNode, k)), "0 0 0")) };
						local scaleMax = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#scaleMax", keyNode, k)), "0 0 0")) };
						local basePosition = { getTranslation(node) };
						local transAxis = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#transAxis", keyNode, k)), "0 0 0")) };
						local transMax = { Utils.getVectorFromString(Utils.getNoNil(getXMLString(xmlFile, string.format("%s.node(%d)#transMax", keyNode, k)), "0 0 0")) };
						local orientateToWorldY = Utils.getNoNil(getXMLBool(xmlFile, string.format("%s.node(%d)#orientateToWorldY", keyNode, k)), false);
						table.insert(volumeHeight.nodes, {node=node, baseScale=baseScale, scaleAxis=scaleAxis, scaleMax=scaleMax, basePosition=basePosition, transAxis=transAxis, transMax=transMax, orientateToWorldY=orientateToWorldY});
					end;
					k = k + 1;
				end;

				table.insert(fillVolumeHeights, volumeHeight);
				j = j + 1;
			end;
		end;
		
		-- ## WRITE TABLE ##
		local entry = {};
		entry.capacity = capacity;
		entry.allowFillFromAir = allowFillFromAir;
		entry.dynamicMountAttacher = dynamicMountAttacher;
		entry.leftSidewallActive = leftSidewallActive;
		entry.rightSidewallActive = rightSidewallActive;
		entry.components = components;
		entry.componentsCount = table.getn(components);
		entry.coverSheet = coverSheet;
		entry.coverSheetUnused = coverSheetUnused;
		entry.coverSheetScaleNode = coverSheetScaleNode;
		entry.coverSheetScaleAnimCurve = coverSheetScaleAnimCurve;
		entry.fillTypes = fillTypes;
		entry.fillPlanes = fillPlanes;
		entry.realCapacityFx = realCapacityFx;
		entry.tipReferencePoints = tipReferencePoints;
		entry.fillVolumes = fillVolumes;
		entry.fillVolumeDeformers = fillVolumeDeformers;
		entry.fillVolumeHeights = fillVolumeHeights;
		entry.fillVolumeHeightRefNodeToFillVolumeHeight = fillVolumeHeightRefNodeToFillVolumeHeight;
		table.insert(self.multiT.body, entry);
		
		i = i + 1;
	end;
	
	-- ## TRAILER TIP REFERENCE POINTS ##
	self.multiT.tipReferencePoints = {};
	for i,j in pairs(self.tipReferencePoints) do
		if self.tipReferencePoints[i] ~= nil then
			self.multiT.tipReferencePoints[i] = self.tipReferencePoints[i];
		end;
	end;
	
	-- ## VARIABLE UNLOAD TIME FROM CURRENT BODY CAPACITY ##
	self.multiT.dischargeEndTime = {};
	self.multiT.dischargeTimeScale = Utils.getNoNil(getXMLString(xmlFile, "vehicle.multiTrailer#dischargeTimeScale"), 1.75);
	for i,j in pairs(self.tipAnimations) do
		if self.tipAnimations[i] ~= nil then
			self.multiT.dischargeEndTime[i] = self.tipAnimations[i].dischargeEndTime;
		end;
	end;
	
	-- ## DYNAMIC MOUNT ATTACHER ##
	if self.dynamicMountAttacherTrigger ~= nil then
		self.multiT.dmat = {};
		local x, y, z = getTranslation(self.dynamicMountAttacherTrigger.triggerNode);
		self.multiT.dmat[1] = x;
		self.multiT.dmat[2] = y;
		self.multiT.dmat[3] = z;
	end;
	
	-- ## UNLOAD ATTACHED OBJECTS ##
	self.multiT.unloadAnimation = getXMLString(xmlFile, "vehicle.multiTrailer#unloadAnimation");
	self.multiT.unloading = false;
	self.isObjectMounted = false;
	
	-- ## OTHER ##
	self.multiT.bodyCount = table.getn(self.multiT.body);
	self.multiT.rainWarning = 0;
	self.multiT.notOpenWarning = 0;
	
	if self.multiT.activeBody > self.multiT.bodyCount then
		self.multiT.activeBody = 1;
	end;
end;

function multiTrailer:postLoad(xmlFile)
	if self.fillPlanes == nil then
		self.fillPlanes = {};
	end;
	if self.spareFillPlanes == nil then
		self.spareFillPlanes = {};
	end;
	
	self.defaultFillPlanes = {};
	for fillType, _ in pairs(self.fillPlanes) do
		self.defaultFillPlanes[fillType] = self.fillPlanes[fillType];
		if self.debugRendering then
			print("INFO multiTrailer: defaultPlane for '" .. fillType .. "' is added from <fillPlanes>");
		end;
	end;
	
	for fillType, _ in pairs(self.spareFillPlanes) do
		if self.defaultFillPlane == nil then
			self.defaultFillPlane = self.spareFillPlanes[fillType];
		end;
		if self.defaultFillPlanes[fillType] == nil then
			self.defaultFillPlanes[fillType] = self.spareFillPlanes[fillType];
			if self.debugRendering then
				print("INFO multiTrailer: defaultPlane for '" .. fillType .. "' is added from spareFillPlanes");
			end;
		end;
	end;
end;

-- ### SAVE / LOAD data & Stream data ###
function multiTrailer:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'currentBody="'..tostring(self.multiT.currentBody)..'" cover="'..tostring(self.multiT.coverSheet)..'"';
	if self.multiT.leftSidewall ~= nil then attributes = attributes..' leftSidewall="'..tostring(self.multiT.leftSidewall.isOpen)..'"'; end;
	if self.multiT.rightSidewall ~= nil then attributes = attributes..' rightSidewall="'..tostring(self.multiT.rightSidewall.isOpen)..'"'; end;
	return attributes, nil;
end;

function multiTrailer:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local currentBody = Utils.getNoNil(getXMLInt(xmlFile, key .. "#currentBody"), 1);
		local coverSheet = Utils.getNoNil(getXMLBool(xmlFile, key .. "#cover"), false);
		local fillLevel = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#fillLevel"), 0);
		local fillType = getXMLString(xmlFile, key.."#fillType");
		
		if currentBody > self.multiT.bodyCount then
			currentBody = 1;
		end;
		
		if fillLevel ~= nil and fillLevel ~= self.fillLevel and fillType ~= nil then
			local fillTypeInt = Fillable.fillTypeNameToInt[fillType];
			if fillTypeInt ~= nil and fillTypeInt ~= Fillable.FILLTYPE_UNKNOWN then
				self:changeCurrentBody(currentBody, fillLevel, fillTypeInt, true);
			else
				self:changeCurrentBody(currentBody, nil, nil, true);
			end;
		else
			self:changeCurrentBody(currentBody, nil, nil, true);
		end;
		
		self:setCoverSheet(coverSheet, true);
		
		if self.multiT.leftSidewall ~= nil then
			local isOpen = Utils.getNoNil(getXMLBool(xmlFile, key .. "#leftSidewall"), false);
			if self.multiT.leftSidewall.isOpen ~= isOpen then self:setOpenableSidewalls("left", isOpen, true); end;
		end;
		
		if self.multiT.rightSidewall ~= nil then
			local isOpen = Utils.getNoNil(getXMLBool(xmlFile, key .. "#rightSidewall"), false);
			if self.multiT.rightSidewall.isOpen ~= isOpen then self:setOpenableSidewalls("right", isOpen, true); end;
		end;
	else
		self:changeCurrentBody(self.multiT.activeBody, nil, nil, true);
	end;
	return BaseMission.VEHICLE_LOAD_OK;
end;

function multiTrailer:readStream(streamId, connection)
	if connection:getIsServer() then
		local currentBody = streamReadInt8(streamId);
		local fillLevel = streamReadFloat32(streamId);
		local fillType = streamReadUIntN(streamId, Fillable.sendNumBits);
		self:changeCurrentBody(currentBody, fillLevel, fillType, true);
		
		self:setCoverSheet(streamReadBool(streamId), true);
		self:setUnloadingObjects(streamReadBool(streamId), true);
		if self.multiT.leftSidewall ~= nil then self:setOpenableSidewalls("left", streamReadBool(streamId), true); end;
		if self.multiT.rightSidewall ~= nil then self:setOpenableSidewalls("right", streamReadBool(streamId), true); end;
	end;
end;

function multiTrailer:writeStream(streamId, connection)
	if not connection:getIsServer() then
		streamWriteInt8(streamId, self.multiT.currentBody);
		streamWriteFloat32(streamId, self.fillLevel);
		streamWriteUIntN(streamId, self.currentFillType, Fillable.sendNumBits);
		
		streamWriteBool(streamId, self.multiT.coverSheet);
		streamWriteBool(streamId, self.multiT.unloading);
		if self.multiT.leftSidewall ~= nil then streamWriteBool(streamId, self.multiT.leftSidewall.isOpen); end;
		if self.multiT.rightSidewall ~= nil then streamWriteBool(streamId, self.multiT.rightSidewall.isOpen); end;
	end;
end;

function multiTrailer:readUpdateStream(streamId, timestamp, connection)
	if connection:getIsServer() then
		self.isObjectMounted = streamReadBool(streamId);
	end;
end;

function multiTrailer:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection:getIsServer() then
		streamWriteBool(streamId, self.isObjectMounted);
	end;
end;

-- ### UPDATE ###
function multiTrailer:update(dt)
	if self.multiT.firstLoad and self.multiT.currentBody == nil then
		self:changeCurrentBody(self.multiT.activeBody)
		self.multiT.firstLoad = false;
	end;
	
	-- ## CHANGE CURRENT BODY ##
	if self:isActiveForInput() then
		if self.multiT.isChangeable then
			if InputBinding.hasEvent(InputBinding.MT_changeCurrentBody) then
				local currentBody = self.multiT.currentBody + 1;
				if currentBody > self.multiT.bodyCount then
					currentBody = 1;
				end;
				self:changeCurrentBody(currentBody);
			end;
		end;
	end;
	
	-- ## OVERRIDE BRAKES WITH MAXIMAL LOAD ##
	if self.multiT.realMaxLoad > 0 and self.isRealistic then
		if self.realFillVolume ~= nil and self.realFillVolume > 0 and self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN then
			local fillTypeName = Fillable.fillTypeIntToName[self.currentFillType];
			local found, density, pricePerKg = RealisticUtils.getFillTypeInfosV2(fillTypeName);
			local fx = 0.001 * density;
			self.multiT.realCropLoad = self.realFillVolume * fx;
			if self.drive then
				self.realBaseCapacity = math.min(math.floor(self.multiT.realMaxLoad / fx), self.multiT.body[self.multiT.currentBody].capacity);
			else
				self.realBaseCapacity = self.multiT.body[self.multiT.currentBody].capacity;
			end;
		elseif self.multiT.balesMass > 0 then
			self.multiT.realCropLoad = self.multiT.balesMass;
		else
			self.multiT.realCropLoad = 0;
		end;
		
		if self.multiT.realCropLoad > self.multiT.realMaxLoad then
			local decrease = ((self.multiT.realCropLoad - self.multiT.realMaxLoad) * 0.11)
			if self.attacherVehicle ~= nil and self.realBrakePedalPosition > 0 then
				self.multiT.brakeDecrease = math.min(decrease, self.multiT.brakeDecrease + (decrease * 0.01));
			else
				self.multiT.brakeDecrease = math.max(0, self.multiT.brakeDecrease - 0.003);
			end;
			 decrease = decrease + self.multiT.brakeDecrease;
			for _, wheel in pairs(self.wheels) do
				wheel.brakeRatio = math.max(0, wheel.brakeRatioBackup - decrease);
			end;
		else
			for _, wheel in pairs(self.wheels) do
				wheel.brakeRatio = wheel.brakeRatioBackup;
			end;
		end;
		
		if g_gui.currentGui == nil and self.multiT.realCropLoad > 0 and g_currentMission.showHelpText then
			local x, y, z = getWorldTranslation(self.components[1].node);
			local x1, y1, z1 = 0, 0, 0;
			if g_currentMission.player ~= nil then
				if g_currentMission.controlledVehicle ~= nil then
					x1, y1, z1 = getWorldTranslation(g_currentMission.controlledVehicle.cameras[g_currentMission.controlledVehicle.camIndex].cameraNode);
				else
					x1, y1, z1 = getWorldTranslation(g_currentMission.player.rootNode);
				end;
				if Utils.vector3Length(x-x1, y-y1, z-z1) < 30 then
					if self.nicknameRenderNodeOffset ~= nil and self.nicknameRenderNodeOffset[1] ~= nil then
						local xn, yn, zn = getWorldTranslation(self.nicknameRenderNode);
						x = xn + self.nicknameRenderNodeOffset[1];
						y = yn + self.nicknameRenderNodeOffset[2];
						z = zn + self.nicknameRenderNodeOffset[3];
					else
						y = y + 3.1;
					end;
					local sx,sy,sz = project(x,y,z);
					if sz <= 1 then
						setTextAlignment(RenderText.ALIGN_CENTER);
						setTextBold(false);
						setTextColor(0.0, 0.0, 0.0, 0.75);
						renderText(sx, sy + 0.05, 0.02, tostring(0.1 * math.floor(self.multiT.realCropLoad * 10)).." t");
						if self.multiT.realCropLoad > self.multiT.realMaxLoad then
							setTextColor(1, 0, 0, 1);
						elseif self.multiT.realCropLoad > (self.multiT.realMaxLoad * 0.9) then
							setTextColor(1, 0.89, 0.267, 1);
						else
							setTextColor(0.5, 1.0, 0.5, 1.0);
						end;
						renderText(sx, sy + 0.05, 0.02, tostring(0.1 * math.floor(self.multiT.realCropLoad * 10)).." t");
						setTextAlignment(RenderText.ALIGN_LEFT);
					end;
				end;
			end;
		end;
	end;
	
	-- ## OPENABLE SIDEWALLS ##
	if self.multiT.body[self.multiT.currentBody].leftSidewallActive and self.multiT.leftSidewall ~= nil and not self.multiT.coverSheet then
		if self.fillLevel <= 0 and self.multiT.leftSidewall.inRange then
			local animTime = self.animations[self.multiT.leftSidewall.animation].currentTime / self.animations[self.multiT.leftSidewall.animation].duration;
			if animTime == 0 and not self.multiT.leftSidewall.isOpen then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_sidewall"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA3);
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
					self:setOpenableSidewalls("left", true);
				end;
			elseif animTime == 1 and self.multiT.leftSidewall.isOpen then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_sidewall"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA3);
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
					self:setOpenableSidewalls("left", false);
				end;
			end;
		end;
	end;
	if self.multiT.body[self.multiT.currentBody].rightSidewallActive and self.multiT.rightSidewall ~= nil and not self.multiT.coverSheet then
		if self.fillLevel <= 0 and self.multiT.rightSidewall.inRange then
			local animTime = self.animations[self.multiT.rightSidewall.animation].currentTime / self.animations[self.multiT.rightSidewall.animation].duration;
			if animTime == 0 and not self.multiT.rightSidewall.isOpen then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_sidewall"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA3);
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
					self:setOpenableSidewalls("right", true);
				end;
			elseif animTime == 1 and self.multiT.rightSidewall.isOpen then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_sidewall"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA3);
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA3) then
					self:setOpenableSidewalls("right", false);
				end;
			end;
		end;
	end;
	if self.multiT.sidewallsChain ~= nil then
		if self.isObjectMounted or (self.multiT.leftSidewall ~= nil and (self.animations[self.multiT.leftSidewall.animation].currentTime / self.animations[self.multiT.leftSidewall.animation].duration) ~= 0) or (self.multiT.rightSidewall ~= nil and (self.animations[self.multiT.rightSidewall.animation].currentTime / self.animations[self.multiT.rightSidewall.animation].duration) ~= 0) then
			if getVisibility(self.multiT.sidewallsChain) then setVisibility(self.multiT.sidewallsChain, false); end;
		else
			if not getVisibility(self.multiT.sidewallsChain) then setVisibility(self.multiT.sidewallsChain, true); end;
		end;
	end;
	
	-- ## SET COVERING SHEET ##
	if self.multiT.body[self.multiT.currentBody].coverSheet ~= nil then
		if (self.multiT.leftSidewall == nil or (self.multiT.leftSidewall ~= nil and not self.multiT.leftSidewall.isOpen)) and (self.multiT.rightSidewall == nil or (self.multiT.rightSidewall ~= nil and not self.multiT.rightSidewall.isOpen)) and not self.drive and not self.isObjectMounted then
			if g_currentMission.environment.lastRainScale > 0 and not self.multiT.coverSheet then
				self:setCoverSheet(true);
			end;
			if self.multiT.coverSheetInRange then
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
					if g_currentMission.environment.lastRainScale == 0 then
						self:setCoverSheet(not self.multiT.coverSheet);
					else
						self.multiT.rainWarning = self.time + 1000;
					end;
				end;
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_coverSheet"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA2);
			end;
			if self.multiT.rainWarning > self.time then
				g_currentMission:addWarning(string.format(g_i18n:getText("MT_rainWarning")), 0.018, 0.033);
			end;
			
			if self.multiT.coverSheet and self.multiT.body[self.multiT.currentBody].coverSheetScaleNode ~= nil then
				if self.fillLevel ~= self.multiT.fillLevel then
					self.multiT.fillLevel = self.fillLevel;
					local t = self.fillLevel / self.multiT.body[self.multiT.currentBody].capacity;
					local x,y,z, rx,ry,rz, sx,sy,sz = self.multiT.body[self.multiT.currentBody].coverSheetScaleAnimCurve:get(t);
					setTranslation(self.multiT.body[self.multiT.currentBody].coverSheetScaleNode, x, y, z);
					setRotation(self.multiT.body[self.multiT.currentBody].coverSheetScaleNode, rx, ry, rz);
					setScale(self.multiT.body[self.multiT.currentBody].coverSheetScaleNode, sx, sy, sz);
				end;
			end;
		end;
		if self.multiT.coverSheet and (self.tipState ~= Trailer.TIPSTATE_CLOSED or (self.multiT.leftSidewall ~= nil and self.multiT.leftSidewall.isOpen) or (self.multiT.rightSidewall ~= nil and self.multiT.rightSidewall.isOpen)) then
			self:setCoverSheet(false);
		end;
	end;
	
	-- ## SET UNLOAD TIME FROM CURRENT BODY CAPACITY ##
	for i,j in pairs(self.tipAnimations) do
		if self.tipAnimations[i] ~= nil then
			if not self.isRealistic or not self.drive then
				self.tipAnimations[i].dischargeEndTime = self.capacity * self.multiT.dischargeTimeScale;
			else
				self.tipAnimations[i].dischargeEndTime = self.multiT.dischargeEndTime[i];
			end;
		end;
	end;
	
	-- ## SET REAL CAPACITY FROM MULTIPLIER ##
	if not self.isRealistic and self.multiT.realCapacity then
		if self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN and self.multiT.body[self.multiT.currentBody].realCapacityFx[self.currentFillType] ~= nil then
			if self.multiT.body[self.multiT.currentBody].realCapacityFx[self.currentFillType] ~= 1 and self.capacity == self.multiT.body[self.multiT.currentBody].capacity then
				self.capacity = tonumber(self.multiT.body[self.multiT.currentBody].capacity * self.multiT.body[self.multiT.currentBody].realCapacityFx[self.currentFillType]);
			end;
		end;
	end;
	
	-- ## UNLOAD ATTACHED OBJECTS ##
	if self.multiT.unloadAnimation ~= nil and self.animations ~= nil and self.animations[self.multiT.unloadAnimation] ~= nil then
		if self:getIsActiveForInput() and self.multiT.body[self.multiT.currentBody].dynamicMountAttacher then
			if not self.multiT.unloading and self.isObjectMounted and not self:getIsAnimationPlaying(self.multiT.unloadAnimation) then
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
					if self.multiT.rightSidewall == nil or (self.multiT.rightSidewall ~= nil and self.multiT.rightSidewall.isOpen) then
						self:setUnloadingObjects(true);
					else
						self.multiT.notOpenWarning = self.time + 1000;
					end;
				end;
				if self.multiT.notOpenWarning > self.time then
					g_currentMission:addWarning(string.format(g_i18n:getText("MT_notOpenWarning")), 0.018, 0.033);
				end;
			elseif self.multiT.unloading and not self:getIsAnimationPlaying(self.multiT.unloadAnimation) then
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
					self:setUnloadingObjects(false);
				end;
			end;
		end;
	end;
end;

-- ### UPDATE TICK ###
function multiTrailer:updateTick(dt)
	self.multiT.isChangeable = false;
	
	if self:getIsActive() and self.multiT.numBalesInTrigger ~= nil then
		self.isObjectMounted = self.multiT.numBalesInTrigger > 0;
	end;
	
	if self.attacherJoint ~= nil then
		if self.attacherVehicle ~= nil and self.attacherVehicle.drive ~= nil then
			self.drive = self.attacherVehicle.drive;
		else
			self.drive = nil;
		end;
	end;
	
	if self.multiT.bodyCount > 1 then
		local nextBody = self.multiT.body[1];
		if self.multiT.currentBody < self.multiT.bodyCount then
			nextBody = self.multiT.body[self.multiT.currentBody + 1];
		end;
		if not self.multiT.onlyRepair or (self.multiT.onlyRepair and self.isInRepairTrigger) then
			if self.tipState == Trailer.TIPSTATE_CLOSED and not self.multiT.unloading and not self.isObjectMounted and (self.fillLevel <= 0 or (self.fillLevel < nextBody.capacity and nextBody.fillTypes[self.currentFillType])) then
				if (self.multiT.leftSidewall == nil or (self.multiT.leftSidewall ~= nil and not self.multiT.leftSidewall.isOpen)) and (self.multiT.rightSidewall == nil or (self.multiT.rightSidewall ~= nil and not self.multiT.rightSidewall.isOpen)) then
					self.multiT.isChangeable = true
				end;
			end;
		end;
	end;
end;

-- ### DRAW Help ###
function multiTrailer:draw()
	if self.multiT.isChangeable then
		g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_currentBodyDraw"), self.multiT.currentBody, self.multiT.bodyCount), InputBinding.MT_changeCurrentBody);
	end;
	if self.multiT.unloadAnimation ~= nil and self.animations ~= nil and self.animations[self.multiT.unloadAnimation] ~= nil then
		if self:getIsActiveForInput() and self.multiT.body[self.multiT.currentBody].dynamicMountAttacher then
			if not self.multiT.unloading and self.isObjectMounted and not self:getIsAnimationPlaying(self.multiT.unloadAnimation) then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_unloadStart"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA2);
			elseif self.multiT.unloading and not self:getIsAnimationPlaying(self.multiT.unloadAnimation) then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("MT_unloadStop"), self.typeDesc), InputBinding.IMPLEMENT_EXTRA2);
			end;
		end;
	end;
end;

-- ### Change Current Body ###
function multiTrailer:changeCurrentBody(state, fillLevel, fillType, noEventSend)
	multiTrailerEvent.sendEvent(self, state, noEventSend);
	self.multiT.currentBody = state;
	
	for _, body in ipairs(self.multiT.body) do
		for _, component in ipairs(body.components) do
			setVisibility(component, false);
		end;
	end;
	
	if self.multiT.body[state] ~= nil then
		for _, component in ipairs(self.multiT.body[state].components) do
			setVisibility(component, true);
		end;
	end;
		
	for i, _ in pairs(self.fillTypes) do
		if self.multiT.body[state].fillTypes[i] == true then
			self.fillTypes[i] = true;
		else
			self.fillTypes[i] = false;
		end;
	end;
	
	for typeFill, _ in pairs(self.multiT.body[state].fillPlanes) do
		self.fillPlanes[typeFill] = self.multiT.body[state].fillPlanes[typeFill];
	end;
	for typeFill, fillPlane in pairs(self.fillPlanes) do
		local fillType = self.fillTypes[Fillable.fillTypeNameToInt[typeFill]];
		if not fillType then
			for _, node in ipairs(fillPlane.nodes) do
				setVisibility(node.node, false);
			end;
			self.fillPlanes[typeFill] = nil;
			if self.debugRendering then
				if fillType == false then
					print("INFO multiTrailer: Plane for '" .. typeFill .. "' is not used for body '" .. tostring(state) .. "' and remove");
				elseif fillType == nil then
					print("INFO multiTrailer: Plane for '" .. typeFill .. "' is not used in this map and remove");
				end;
			end;
		end;
	end;
	for typeFill, _ in pairs(self.fillTypes) do
		local typeFillName = Fillable.fillTypeIntToName[typeFill];
		if self.fillTypes[typeFill] and self.fillPlanes[typeFillName] == nil then
			self.fillPlanes[typeFillName] = self.defaultFillPlanes[typeFillName];
			if self.debugRendering then
				print("INFO multiTrailer: Plane for '" .. typeFillName .. "' is not defined for body '" .. tostring(state) .. "' and added from defaultFillPlanes");
			end;
		end;
	end;
	
	for i, j in pairs(self.multiT.tipReferencePoints) do
		self.tipReferencePoints[i] = self.multiT.tipReferencePoints[i];
	end;
	for i, j in pairs(self.tipReferencePoints) do
		if not self.multiT.body[state].tipReferencePoints[i] then
			table.remove(self.tipReferencePoints, i);
		end;
	end;
	
	self.allowFillFromAir = self.multiT.body[state].allowFillFromAir;
	self.capacity = self.multiT.body[state].capacity;
	
	-- ## MORE REALISTIC ##
	if self.isRealistic then
		self.realFillVolume = self.fillLevel;
		self.realBaseCapacity = self.multiT.body[state].capacity;
		for i, j in pairs(self.fillTypes) do
			if self.fillTypes[i] then
				self.realCapacityFx[i] = Utils.getNoNil(self.multiT.body[state].realCapacityFx[i], 1);
			end;
		end;
	end;
	
	-- ## ONLY FOR USE IN LS15 ##
	if self.fillVolumes ~= nil then
		for _, body in ipairs(self.multiT.body) do
			for _, fillVolume in pairs(body.fillVolumes) do
				if fillVolume.volume ~= nil then
					setVisibility(fillVolume.volume, false);
				end;
			end;
		end;
		
		for i, _ in pairs(self.multiT.body[state].fillVolumes) do
			self.fillVolumes[i] = self.multiT.body[state].fillVolumes[i];
		end;
		
		for node, _ in pairs(self.fillVolumeDeformers) do
			if self.multiT.body[state].fillVolumeDeformers[node] == nil then
				table.remove(self.fillVolumeDeformers, node);
			end;
		end;
		for node, _ in pairs(self.multiT.body[state].fillVolumeDeformers) do
			self.fillVolumeDeformers[node] = self.multiT.body[state].fillVolumeDeformers[node];
		end;
		
		if self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN then
			local usedFillType = self.currentFillType;
			if self.forcedFillPlaneType ~= nil then
				usedFillType = self.forcedFillPlaneType;
			end;
			materialId, isDefault = Fillable.getFillMaterial(usedFillType, Fillable.MATERIAL_TYPE_FILLPLANE, 1);
			for _, fillVolume in pairs(self.fillVolumes) do
				if isDefault and fillVolume.defaultFillType ~= nil then
					materialId = Fillable.getFillMaterial(fillVolume.defaultFillType, Fillable.MATERIAL_TYPE_FILLPLANE, 1);
				end;
				if materialId ~= nil then
					setMaterial(fillVolume.volume, materialId, 0);
				end;
				local delta = math.min(Utils.getNoNil(fillLevel, self.fillLevel), self.capacity);
				local x,y,z = localToWorld(fillVolume.volume, 0,100,0);
				local d1x,d1y,d1z = localDirectionToWorld(fillVolume.volume, 5,0,0);
				local d2x,d2y,d2z = localDirectionToWorld(fillVolume.volume, 0,0,5);
				x = x - (d1x+d2x)/2;
				y = y - (d1y+d2y)/2;
				z = z - (d1z+d2z)/2;
				fillPlaneAdd(fillVolume.volume, delta, x,y,z, d1x,d1y,d1z, d2x,d2y,d2z);
				self:setFillLevel(self.fillLevel, usedFillType);
			end;
		end;
		
		for i, _ in pairs(self.fillVolumeHeights) do
			if self.multiT.body[state].fillVolumeHeights[i] == nil then
				table.remove(self.fillVolumeHeights, i);
			end;
		end;
		for i, _ in pairs(self.multiT.body[state].fillVolumeHeights) do
			self.fillVolumeHeights[i] = self.multiT.body[state].fillVolumeHeights[i];
		end;
		
		for node, _ in pairs(self.fillVolumeHeightRefNodeToFillVolumeHeight) do
			if self.multiT.body[state].fillVolumeHeightRefNodeToFillVolumeHeight[node] == nil then
				table.remove(self.fillVolumeHeightRefNodeToFillVolumeHeight, node);
			end;
		end;
		for node, _ in pairs(self.multiT.body[state].fillVolumeHeightRefNodeToFillVolumeHeight) do
			self.fillVolumeHeightRefNodeToFillVolumeHeight[node] = self.multiT.body[state].fillVolumeHeightRefNodeToFillVolumeHeight[node];
		end;
	end;
	
	-- ## SET FILLLEVEL ##
	if fillLevel ~= nil and fillType ~= nil then
		self:setFillLevel(fillLevel, fillType);
	end;
	
	-- ## SET DYNAMIC MOUNT ATTACHER ##
	if self.multiT.dmat ~= nil then
		if self.multiT.body[state].dynamicMountAttacher then
			setTranslation(self.dynamicMountAttacherTrigger.triggerNode, self.multiT.dmat[1], self.multiT.dmat[2], self.multiT.dmat[3]);
		else
			setTranslation(self.dynamicMountAttacherTrigger.triggerNode, self.multiT.dmat[1], self.multiT.dmat[2] - 5, self.multiT.dmat[3]);
		end;
	end;
	
	self:setCoverSheet(false);
end;

function multiTrailer:coverSheetTrigger(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter or onLeave then
		if g_currentMission.player ~= nil and otherActorId == g_currentMission.player.rootNode then
			if onEnter and self.multiT.body[self.multiT.currentBody].coverSheet ~= nil then
				self.multiT.coverSheetInRange = true;
			else
				self.multiT.coverSheetInRange = false;
			end;
		end;
	end;
end;

function multiTrailer:leftTrigger(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter or onLeave then
		if g_currentMission.player ~= nil and otherActorId == g_currentMission.player.rootNode then
			if onEnter then
				self.multiT.leftSidewall.inRange = true;
			else
				self.multiT.leftSidewall.inRange = false;
			end;
		end;
	end;
end;

function multiTrailer:rightTrigger(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter or onLeave then
		if g_currentMission.player ~= nil and otherActorId == g_currentMission.player.rootNode then
			if onEnter then
				self.multiT.rightSidewall.inRange = true;
			else
				self.multiT.rightSidewall.inRange = false;
			end;
		end;
	end;
end;

function multiTrailer:realisticBaleSearchTrigger(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	if onEnter then
		local object = g_currentMission:getNodeObject(otherActorId);
		if object ~= nil then
			if object:isa(Bale) then
				self.multiT.numBalesInTrigger = self.multiT.numBalesInTrigger + 1;
				if self.isRealistic and object.isRealistic then
					self.multiT.balesMass = self.multiT.balesMass + object.realCurrentMass;
				end;
			end;
		end;
	elseif onLeave then
		local object = g_currentMission:getNodeObject(otherActorId);
		if object ~= nil then
			if object:isa(Bale) then
				self.multiT.numBalesInTrigger = self.multiT.numBalesInTrigger - 1;
				if self.isRealistic and object.isRealistic then
					self.multiT.balesMass = self.multiT.balesMass - object.realCurrentMass;
				end;
			end;
		end;
	end;
end;

function multiTrailer:setCoverSheet(coverSheet, noEventSend)
	SetCoverSheetEvent.sendEvent(self, coverSheet, noEventSend);
	self.multiT.coverSheet = coverSheet;
	if coverSheet then
		self.allowFillFromAir = false;
		if self.multiT.body[self.multiT.currentBody].coverSheetUnused ~= nil then
			setVisibility(self.multiT.body[self.multiT.currentBody].coverSheetUnused, false);
		end;
		setVisibility(self.multiT.body[self.multiT.currentBody].coverSheet, true);
	else
		self.allowFillFromAir = self.multiT.body[self.multiT.currentBody].allowFillFromAir;
		for _, body in ipairs(self.multiT.body) do
			if body.coverSheet ~= nil then
				setVisibility(body.coverSheet, false);
			end;
			if body.coverSheetUnused ~= nil then
				setVisibility(body.coverSheetUnused, false);
			end;
		end;
		if self.multiT.body[self.multiT.currentBody].coverSheetUnused ~= nil then
			setVisibility(self.multiT.body[self.multiT.currentBody].coverSheetUnused, true);
		end;
	end;
end;

function multiTrailer:setOpenableSidewalls(side, isOpen, noEventSend)
	SetOpenableSidewallsEvent.sendEvent(self, side, isOpen, noEventSend);
	local direction = 0;
	if isOpen then
		direction = 1;
		self.allowFillFromAir = false;
		self.capacity = 0;
		self.realBaseCapacity = 0;
	else
		direction = -1;
		self.allowFillFromAir = self.multiT.body[self.multiT.currentBody].allowFillFromAir;
		self.capacity = self.multiT.body[self.multiT.currentBody].capacity;
		self.realBaseCapacity = self.multiT.body[self.multiT.currentBody].capacity;
	end;
	
	if side == "left" then
		self.multiT.leftSidewall.isOpen = isOpen;
		self:playAnimation(self.multiT.leftSidewall.animation, direction, nil, true);
	elseif side == "right" then
		self.multiT.rightSidewall.isOpen = isOpen;
		self:playAnimation(self.multiT.rightSidewall.animation, direction, nil, true);
	end;
end;

function multiTrailer:setUnloadingObjects(unloading, noEventSend)
	SetUnloadingObjectsEvent.sendEvent(self, unloading, noEventSend)
	self.multiT.unloading = unloading;
	if unloading then
		if self.multiT.dmat ~= nil then
			if self.isServer then
				for object,_ in pairs(self.pendingDynamicMountObjects) do
					object:unmountDynamic();
				end;
				setTranslation(self.dynamicMountAttacherTrigger.triggerNode, self.multiT.dmat[1], self.multiT.dmat[2] - 5, self.multiT.dmat[3]);
			end;
		end;
		self:playAnimation(self.multiT.unloadAnimation, 1, nil, true);
	else
		if self.multiT.dmat ~= nil then
			if self.isServer then
				setTranslation(self.dynamicMountAttacherTrigger.triggerNode, self.multiT.dmat[1], self.multiT.dmat[2], self.multiT.dmat[3]);
			end;
		end;
		self:playAnimation(self.multiT.unloadAnimation, -1, nil, true);
	end;
end;

function multiTrailer:isActiveForInput(onlyTrueIfSelected)
	if g_gui.currentGui ~= nil or g_currentMission.isPlayerFrozen then
		return false;
	end;
	if self.isEntered then
		if onlyTrueIfSelected == nil or onlyTrueIfSelected then
			return self.selectedImplement == nil;
		else
			return true;
		end;
	end;
	if self.attacherVehicle ~= nil then
		if self.attacherVehicle.isActiveForInput ~= nil then
			if onlyTrueIfSelected == nil or onlyTrueIfSelected then
				return self.isSelected and self.attacherVehicle:isActiveForInput(false);
			else
				return self.attacherVehicle:isActiveForInput(false);
			end;
		end;
		if onlyTrueIfSelected == nil or onlyTrueIfSelected then
			return self.isSelected and self.attacherVehicle:getIsActiveForInput(false);
		else
			return self.attacherVehicle:getIsActiveForInput(false);
		end;
	end;
	return false;
end;

function multiTrailer:onAttach(attacherVehicle)
end;

function multiTrailer:onDetach()
end;

function multiTrailer:delete()
	if self.multiT.coverSheetTriggerNode ~= nil then
		removeTrigger(self.multiT.coverSheetTriggerNode);
	end;
	if self.multiT.realisticBaleSearchTriggerNode ~= nil then
		removeTrigger(self.multiT.realisticBaleSearchTriggerNode);
	end;
	if self.multiT.leftTriggerNode ~= nil then
		removeTrigger(self.multiT.leftTriggerNode);
	end;
	if self.multiT.rightTriggerNode ~= nil then
		removeTrigger(self.multiT.rightTriggerNode);
	end;
	
	for _, body in ipairs(self.multiT.body) do
		for _, fillVolume in pairs(body.fillVolumes) do
			if fillVolume.volume ~= nil then
				delete(fillVolume.volume);
			end;
			fillVolume.volume = nil;
		end;
	end;
end;

function multiTrailer:mouseEvent(posX, posY, isDown, isUp, button)
end;

function multiTrailer:keyEvent(unicode, sym, modifier, isDown)
end;